@@ -1,33 +1,10 @@
 #region Copyright notice and license
 // Protocol Buffers - Google's data interchange format
 // Copyright 2008 Google Inc.  All rights reserved.
-// https://developers.google.com/protocol-buffers/
 //
-// Redistribution and use in source and binary forms, with or without
-// modification, are permitted provided that the following conditions are
-// met:
-//
-//     * Redistributions of source code must retain the above copyright
-// notice, this list of conditions and the following disclaimer.
-//     * Redistributions in binary form must reproduce the above
-// copyright notice, this list of conditions and the following disclaimer
-// in the documentation and/or other materials provided with the
-// distribution.
-//     * Neither the name of Google Inc. nor the names of its
-// contributors may be used to endorse or promote products derived from
-// this software without specific prior written permission.
-//
-// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+// Use of this source code is governed by a BSD-style
+// license that can be found in the LICENSE file or at
+// https://developers.google.com/open-source/licenses/bsd
 #endregion
 
 using System;
@@ -630,6 +607,18 @@ namespace Google.Protobuf
             Assert.AreEqual(expectedJson, JsonFormatter.Default.Format(populated));
         }
 
+        // See See https://github.com/protocolbuffers/protobuf/issues/11987
+        [Test]
+        public void JsonNamePriority()
+        {
+            // This tests both the formatter and the parser, but the issue was when parsing.
+            var original = new Issue11987Message { A = 10, B = 20, C = 30 };
+            var json = JsonFormatter.Default.Format(original);
+            AssertJson("{ 'b': 10, 'a': 20, 'd': 30 }", json);
+            var parsed = Issue11987Message.Parser.ParseJson(json);
+            Assert.AreEqual(original, parsed);
+        }
+
         // Sanity tests for WriteValue. Not particularly comprehensive, as it's all covered above already,
         // as FormatMessage uses WriteValue.
 
@@ -854,6 +843,56 @@ namespace Google.Protobuf
             AssertWriteValue(value, expectedJson, JsonFormatter.Settings.Default.WithIndentation());
         }
 
+        [Test]
+        public void WriteValueWithIndentation_MapWithNested()
+        {
+            var value = new TestMap
+            {
+                MapInt32ForeignMessage =
+                {
+                    { 1, new ForeignMessage { C = 1 } },
+                    { 2, new ForeignMessage { C = 2 } },
+                },
+            };
+
+            const string expectedJson = @"
+{
+  'mapInt32ForeignMessage': {
+    '1': {
+      'c': 1
+    },
+    '2': {
+      'c': 2
+    }
+  }
+}";
+
+            AssertWriteValue(value, expectedJson, JsonFormatter.Settings.Default.WithIndentation());
+        }
+
+        [Test]
+        public void WriteValueWithIndentation_MapWithEmptyNested()
+        {
+            var value = new TestMap
+            {
+                MapInt32ForeignMessage =
+                {
+                    { 1, new ForeignMessage() },
+                    { 2, new ForeignMessage() },
+                },
+            };
+
+            const string expectedJson = @"
+{
+  'mapInt32ForeignMessage': {
+    '1': {},
+    '2': {}
+  }
+}";
+
+            AssertWriteValue(value, expectedJson, JsonFormatter.Settings.Default.WithIndentation());
+        }
+
         [Test]
         public void WriteValueWithIndentation_List()
         {
@@ -861,6 +900,42 @@ namespace Google.Protobuf
             AssertWriteValue(value, "[\n  1,\n  2,\n  3\n]", JsonFormatter.Settings.Default.WithIndentation());
         }
 
+        [Test]
+        public void WriteValueWithIndentation_Any()
+        {
+            var registry = TypeRegistry.FromMessages(ForeignMessage.Descriptor);
+            var formatter = JsonFormatter.Settings.Default.WithIndentation().WithTypeRegistry(registry);
+
+            var nestedMessage = new ForeignMessage { C = 1 };
+            var value = Any.Pack(nestedMessage);
+            const string expectedJson = @"
+{
+  '@type': 'type.googleapis.com/protobuf_unittest3.ForeignMessage',
+  'c': 1
+}";
+
+            AssertWriteValue(value, expectedJson, formatter);
+        }
+
+        [Test]
+        public void WriteValueWithIndentation_NestedAny()
+        {
+            var registry = TypeRegistry.FromMessages(ForeignMessage.Descriptor);
+            var formatter = JsonFormatter.Settings.Default.WithIndentation().WithTypeRegistry(registry);
+
+            var nestedMessage = new ForeignMessage { C = 1 };
+            var value = new TestWellKnownTypes { AnyField = Any.Pack(nestedMessage) };
+            const string expectedJson = @"
+{
+  'anyField': {
+    '@type': 'type.googleapis.com/protobuf_unittest3.ForeignMessage',
+    'c': 1
+  }
+}";
+
+            AssertWriteValue(value, expectedJson, formatter);
+        }
+
         [Test]
         public void WriteValueWithIndentation_CustomIndentation()
         {
