@@ -1,32 +1,9 @@
 // Protocol Buffers - Google's data interchange format
 // Copyright 2008 Google Inc.  All rights reserved.
-// https://developers.google.com/protocol-buffers/
 //
-// Redistribution and use in source and binary forms, with or without
-// modification, are permitted provided that the following conditions are
-// met:
-//
-//     * Redistributions of source code must retain the above copyright
-// notice, this list of conditions and the following disclaimer.
-//     * Redistributions in binary form must reproduce the above
-// copyright notice, this list of conditions and the following disclaimer
-// in the documentation and/or other materials provided with the
-// distribution.
-//     * Neither the name of Google Inc. nor the names of its
-// contributors may be used to endorse or promote products derived from
-// this software without specific prior written permission.
-//
-// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+// Use of this source code is governed by a BSD-style
+// license that can be found in the LICENSE file or at
+// https://developers.google.com/open-source/licenses/bsd
 
 #import <Foundation/Foundation.h>
 
@@ -612,6 +589,10 @@ __attribute__((objc_subclassing_restricted))
 - (void)writeGroupNoTag:(int32_t)fieldNumber
                   value:(GPBMessage *)value;
 
+//%PDDM-EXPAND-END _WRITE_DECLS()
+
+// clang-format on
+
 /**
  * Write a GPBUnknownFieldSet for the given field number.
  *
@@ -619,7 +600,9 @@ __attribute__((objc_subclassing_restricted))
  * @param value       The value to write out.
  **/
 - (void)writeUnknownGroup:(int32_t)fieldNumber
-                    value:(GPBUnknownFieldSet *)value;
+                    value:(GPBUnknownFieldSet *)value
+    __attribute__((deprecated("GPBUnknownFieldSet is going away.")));
+
 /**
  * Write an array of GPBUnknownFieldSet for the given field number.
  *
@@ -627,7 +610,9 @@ __attribute__((objc_subclassing_restricted))
  * @param values      The values to write out.
  **/
 - (void)writeUnknownGroupArray:(int32_t)fieldNumber
-                        values:(NSArray<GPBUnknownFieldSet*> *)values;
+                        values:(NSArray<GPBUnknownFieldSet *> *)values
+    __attribute__((deprecated("GPBUnknownFieldSet is going away.")));
+
 /**
  * Write a GPBUnknownFieldSet without any tag (but does write the endGroup tag).
  *
@@ -635,11 +620,8 @@ __attribute__((objc_subclassing_restricted))
  * @param value       The value to write out.
  **/
 - (void)writeUnknownGroupNoTag:(int32_t)fieldNumber
-                         value:(GPBUnknownFieldSet *)value;
-
-//%PDDM-EXPAND-END _WRITE_DECLS()
-
-// clang-format on
+                         value:(GPBUnknownFieldSet *)value
+    __attribute__((deprecated("GPBUnknownFieldSet is going away.")));
 
 /**
 Write a MessageSet extension field to the stream. For historical reasons,
@@ -764,6 +746,5 @@ NS_ASSUME_NONNULL_END
 //%_WRITE_UNPACKABLE_DECLS(Message, GPBMessage)
 //%_WRITE_UNPACKABLE_DECLS(Bytes, NSData)
 //%_WRITE_GROUP_DECLS(Group, GPBMessage)
-//%_WRITE_GROUP_DECLS(UnknownGroup, GPBUnknownFieldSet)
 
 // clang-format on
