@@ -1,32 +1,9 @@
 // Protocol Buffers - Google's data interchange format
 // Copyright 2008 Google Inc.  All rights reserved.
-// https://developers.google.com/protocol-buffers/
 //
-// Redistribution and use in source and binary forms, with or without
-// modification, are permitted provided that the following conditions are
-// met:
-//
-//     * Redistributions of source code must retain the above copyright
-// notice, this list of conditions and the following disclaimer.
-//     * Redistributions in binary form must reproduce the above
-// copyright notice, this list of conditions and the following disclaimer
-// in the documentation and/or other materials provided with the
-// distribution.
-//     * Neither the name of Google Inc. nor the names of its
-// contributors may be used to endorse or promote products derived from
-// this software without specific prior written permission.
-//
-// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+// Use of this source code is governed by a BSD-style
+// license that can be found in the LICENSE file or at
+// https://developers.google.com/open-source/licenses/bsd
 
 #import "GPBUnknownFieldSet_PackagePrivate.h"
 
@@ -36,6 +13,10 @@
 #import "GPBUtilities.h"
 #import "GPBWireFormat.h"
 
+#pragma clang diagnostic push
+#pragma clang diagnostic ignored "-Wdeprecated-declarations"
+#pragma clang diagnostic ignored "-Wdeprecated-implementations"
+
 #pragma mark Helpers
 
 static void checkNumber(int32_t number) {
@@ -230,14 +211,11 @@ static void GPBUnknownFieldSetSerializedSizeAsMessageSet(__unused const void *ke
   NSMutableData *data = [NSMutableData dataWithLength:self.serializedSize];
   GPBCodedOutputStream *output = [[GPBCodedOutputStream alloc] initWithData:data];
   [self writeToCodedOutputStream:output];
+  [output flush];
   [output release];
   return data;
 }
 
-+ (BOOL)isFieldTag:(int32_t)tag {
-  return GPBWireFormatGetTagWireType(tag) != GPBWireFormatEndGroup;
-}
-
 - (void)addField:(GPBUnknownField *)field {
   int32_t number = [field number];
   checkNumber(number);
@@ -294,6 +272,11 @@ static void GPBUnknownFieldSetMergeUnknownFields(__unused const void *key, const
   [[self mutableFieldForNumber:number create:YES] addVarint:value];
 }
 
+- (void)mergeLengthDelimited:(int32_t)fieldNum value:(NSData *)value {
+  checkNumber(fieldNum);
+  [[self mutableFieldForNumber:fieldNum create:YES] addLengthDelimited:value];
+}
+
 - (BOOL)mergeFieldFrom:(int32_t)tag input:(GPBCodedInputStream *)input {
   NSAssert(GPBWireFormatIsValidTag(tag), @"Got passed an invalid tag");
   int32_t number = GPBWireFormatGetTagFieldNumber(tag);
@@ -336,15 +319,6 @@ static void GPBUnknownFieldSetMergeUnknownFields(__unused const void *key, const
   }
 }
 
-- (void)mergeMessageSetMessage:(int32_t)number data:(NSData *)messageData {
-  [[self mutableFieldForNumber:number create:YES] addLengthDelimited:messageData];
-}
-
-- (void)addUnknownMapEntry:(int32_t)fieldNum value:(NSData *)data {
-  GPBUnknownField *field = [self mutableFieldForNumber:fieldNum create:YES];
-  [field addLengthDelimited:data];
-}
-
 - (void)mergeFromCodedInputStream:(GPBCodedInputStream *)input {
   while (YES) {
     int32_t tag = GPBCodedInputStreamReadTag(&input->state_);
@@ -367,3 +341,5 @@ static void GPBUnknownFieldSetMergeUnknownFields(__unused const void *key, const
 #pragma clang diagnostic pop
 
 @end
+
+#pragma clang diagnostic pop  // -Wdeprecated-declarations suppression
