@@ -726,7 +726,7 @@ class GeneratedDescriptorTest(unittest.TestCase):
     excepted_dict['new_key'] = 'new'
     self.assertNotEqual(mapping, excepted_dict)
     self.assertRaises(KeyError, mapping.__getitem__, 'key_error')
-    self.assertRaises(KeyError, mapping.__getitem__, len(mapping) * 2)
+    self.assertRaises(KeyError, mapping.__getitem__, len(mapping) + 1)
     # TODO: Add __repr__ support for DescriptorMapping.
     if api_implementation.Type() == 'cpp':
       self.assertEqual(str(mapping)[0], '<')
@@ -884,6 +884,10 @@ class DescriptorCopyToProtoTest(unittest.TestCase):
         name: 'FOREIGN_BAX'
         number: 32
       >
+      value: <
+        name: 'FOREIGN_LARGE'
+        number: 123456
+      >
       """
 
     self._InternalTestCopyToProto(
