@@ -31,6 +31,7 @@
 #include "absl/strings/string_view.h"
 #include "absl/synchronization/barrier.h"
 #include "absl/utility/utility.h"
+#include "google/protobuf/arena_cleanup.h"
 #include "google/protobuf/arena_test_util.h"
 #include "google/protobuf/descriptor.h"
 #include "google/protobuf/extension_set.h"
@@ -181,21 +182,30 @@ void TestCtorAndDtorTraits(std::vector<absl::string_view> def,
       : std::conditional_t<arena_ctor, ArenaCtorBase, EmptyBase<0>>,
         std::conditional_t<arena_dtor, ArenaDtorBase, EmptyBase<1>>,
         Message {
-    TraitsProber() { actions.push_back("()"); }
-    TraitsProber(const TraitsProber&) { actions.push_back("(const T&)"); }
-    explicit TraitsProber(int) { actions.push_back("(int)"); }
-    explicit TraitsProber(Arena* arena) { actions.push_back("(Arena)"); }
-    TraitsProber(Arena* arena, const TraitsProber&) {
+    TraitsProber() : Message(nullptr, nullptr) { actions.push_back("()"); }
+    TraitsProber(const TraitsProber&) : Message(nullptr, nullptr) {
+      actions.push_back("(const T&)");
+    }
+    explicit TraitsProber(int) : Message(nullptr, nullptr) {
+      actions.push_back("(int)");
+    }
+    explicit TraitsProber(Arena* arena) : Message(nullptr, nullptr) {
+      actions.push_back("(Arena)");
+    }
+    TraitsProber(Arena* arena, const TraitsProber&)
+        : Message(nullptr, nullptr) {
       actions.push_back("(Arena, const T&)");
     }
-    TraitsProber(Arena* arena, int) { actions.push_back("(Arena, int)"); }
-    ~TraitsProber() override { actions.push_back("~()"); }
+    TraitsProber(Arena* arena, int) : Message(nullptr, nullptr) {
+      actions.push_back("(Arena, int)");
+    }
+    ~TraitsProber() { actions.push_back("~()"); }
 
-    TraitsProber* New(Arena*) const final {
+    TraitsProber* New(Arena*) const {
       ABSL_LOG(FATAL);
       return nullptr;
     }
-    const ClassData* GetClassData() const final {
+    const ClassData* GetClassData() const PROTOBUF_FINAL {
       ABSL_LOG(FATAL);
       return nullptr;
     }
@@ -510,11 +520,16 @@ class DispatcherTestProto : public Message {
   using InternalArenaConstructable_ = void;
   using DestructorSkippable_ = void;
   // For the test below to construct.
-  explicit DispatcherTestProto(absl::in_place_t) {}
-  explicit DispatcherTestProto(Arena*) { ABSL_LOG(FATAL); }
-  DispatcherTestProto(Arena*, const DispatcherTestProto&) { ABSL_LOG(FATAL); }
-  DispatcherTestProto* New(Arena*) const final { ABSL_LOG(FATAL); }
-  const ClassData* GetClassData() const final { ABSL_LOG(FATAL); }
+  explicit DispatcherTestProto(absl::in_place_t) : Message(nullptr, nullptr) {}
+  explicit DispatcherTestProto(Arena*) : Message(nullptr, nullptr) {
+    ABSL_LOG(FATAL);
+  }
+  DispatcherTestProto(Arena*, const DispatcherTestProto&)
+      : Message(nullptr, nullptr) {
+    ABSL_LOG(FATAL);
+  }
+  DispatcherTestProto* New(Arena*) const PROTOBUF_FINAL { ABSL_LOG(FATAL); }
+  const ClassData* GetClassData() const PROTOBUF_FINAL { ABSL_LOG(FATAL); }
 };
 // We use a specialization to inject behavior for the test.
 // This test is very intrusive and will have to be fixed if we change the
@@ -878,7 +893,7 @@ TEST(ArenaTest, ReleaseFromArenaMessageUsingReflectionMakesCopy) {
     const Reflection* r = arena_message->GetReflection();
     const FieldDescriptor* f = arena_message->GetDescriptor()->FindFieldByName(
         "optional_nested_message");
-    nested_msg = DownCastToGenerated<TestAllTypes::NestedMessage>(
+    nested_msg = DownCastMessage<TestAllTypes::NestedMessage>(
         r->ReleaseMessage(arena_message, f));
   }
   EXPECT_EQ(42, nested_msg->bb());
@@ -1406,12 +1421,12 @@ TEST(ArenaTest, RepeatedFieldOnArena) {
   // Preallocate an initial arena block to avoid mallocs during hooked region.
   std::vector<char> arena_block(1024 * 1024);
   Arena arena(arena_block.data(), arena_block.size());
+  const size_t initial_allocated_size = arena.SpaceAllocated();
 
   {
-    internal::NoHeapChecker no_heap;
-
-    // Fill some repeated fields on the arena to test for leaks. Also verify no
-    // memory allocations.
+    // Fill some repeated fields on the arena to test for leaks. Also that the
+    // newly allocated memory is approximately the size of the cleanups for the
+    // repeated messages.
     RepeatedField<int32_t> repeated_int32(&arena);
     RepeatedPtrField<TestAllTypes> repeated_message(&arena);
     for (int i = 0; i < 100; i++) {
@@ -1432,10 +1447,14 @@ TEST(ArenaTest, RepeatedFieldOnArena) {
     repeated_message.UnsafeArenaExtractSubrange(0, 5, extracted_messages);
     EXPECT_EQ(&arena, repeated_message.Get(0).GetArena());
     EXPECT_EQ(5, repeated_message.size());
+    // Upper bound of the size of the cleanups of new repeated messages.
+    const size_t upperbound_cleanup_size =
+        2 * 110 * sizeof(internal::cleanup::CleanupNode);
+    EXPECT_GT(initial_allocated_size + upperbound_cleanup_size,
+              arena.SpaceAllocated());
   }
 
-  // Now, outside the scope of the NoHeapChecker, test ExtractSubrange's copying
-  // semantics.
+  // Now test ExtractSubrange's copying semantics.
   {
     RepeatedPtrField<TestAllTypes> repeated_message(&arena);
     for (int i = 0; i < 100; i++) {
@@ -1486,7 +1505,7 @@ TEST(ArenaTest, MutableMessageReflection) {
   const Descriptor* d = message->GetDescriptor();
   const FieldDescriptor* field = d->FindFieldByName("optional_nested_message");
   TestAllTypes::NestedMessage* submessage =
-      DownCastToGenerated<TestAllTypes::NestedMessage>(
+      DownCastMessage<TestAllTypes::NestedMessage>(
           r->MutableMessage(message, field));
   TestAllTypes::NestedMessage* submessage_expected =
       message->mutable_optional_nested_message();
@@ -1496,7 +1515,7 @@ TEST(ArenaTest, MutableMessageReflection) {
 
   const FieldDescriptor* oneof_field =
       d->FindFieldByName("oneof_nested_message");
-  submessage = DownCastToGenerated<TestAllTypes::NestedMessage>(
+  submessage = DownCastMessage<TestAllTypes::NestedMessage>(
       r->MutableMessage(message, oneof_field));
   submessage_expected = message->mutable_oneof_nested_message();
 
@@ -1577,14 +1596,24 @@ TEST(ArenaTest, NoHeapAllocationsTest) {
   Arena arena(options);
 
   {
-
+    // We need to call Arena::Create before NoHeapChecker because the ArenaDtor
+    // allocates a new cleanup chunk.
     TestAllTypes* message = Arena::Create<TestAllTypes>(&arena);
+
+
     FillArenaAwareFields(message);
   }
 
   arena.Reset();
 }
 
+TEST(ArenaTest, ParseCorruptedString) {
+  TestAllTypes message;
+  TestUtil::SetAllFields(&message);
+  TestParseCorruptedString<TestAllTypes, true>(message);
+  TestParseCorruptedString<TestAllTypes, false>(message);
+}
+
 #if PROTOBUF_RTTI
 // Test construction on an arena via generic MessageLite interface. We should be
 // able to successfully deserialize on the arena without incurring heap
@@ -1603,8 +1632,9 @@ TEST(ArenaTest, MessageLiteOnArena) {
   initial_message.SerializeToString(&serialized);
 
   {
-
     MessageLite* generic_message = prototype->New(&arena);
+
+
     EXPECT_TRUE(generic_message != nullptr);
     EXPECT_EQ(&arena, generic_message->GetArena());
     EXPECT_TRUE(generic_message->ParseFromString(serialized));
@@ -1672,6 +1702,23 @@ TEST(ArenaTest, FirstArenaOverhead) {
 }
 
 
+TEST(ArenaTest, StartingBlockSize) {
+  Arena default_arena;
+  EXPECT_EQ(0, default_arena.SpaceAllocated());
+
+  // Allocate something to get starting block size.
+  Arena::CreateArray<char>(&default_arena, 1);
+  ArenaOptions options;
+  // First block size should be the default starting block size.
+  EXPECT_EQ(default_arena.SpaceAllocated(), options.start_block_size);
+
+  // Use a custom starting block size.
+  options.start_block_size *= 2;
+  Arena custom_arena(options);
+  Arena::CreateArray<char>(&custom_arena, 1);
+  EXPECT_EQ(custom_arena.SpaceAllocated(), options.start_block_size);
+}
+
 TEST(ArenaTest, BlockSizeDoubling) {
   Arena arena;
   EXPECT_EQ(0, arena.SpaceUsed());
