@@ -130,6 +130,7 @@ FieldGeneratorBase::FieldGeneratorBase(const FieldDescriptor* field,
       break;
     case FieldDescriptor::CPPTYPE_STRING:
       is_string_ = true;
+      string_type_ = field->options().ctype();
       is_inlined_ = IsStringInlined(field, options);
       is_bytes_ = field->type() == FieldDescriptor::TYPE_BYTES;
       has_default_constexpr_constructor_ = is_repeated_or_map;
@@ -146,6 +147,7 @@ FieldGeneratorBase::FieldGeneratorBase(const FieldDescriptor* field,
   }
 
   has_trivial_zero_default_ = CanInitializeByZeroing(field, options, scc);
+  has_brace_default_assign_ = has_trivial_zero_default_ && !is_lazy_;
 }
 
 void FieldGeneratorBase::GenerateMemberConstexprConstructor(
@@ -227,6 +229,40 @@ void FieldGeneratorBase::GenerateCopyConstructorCode(io::Printer* p) const {
 }
 
 namespace {
+// Use internal types instead of ctype or string_type.
+enum class StringType {
+  kView,
+  kString,
+  kCord,
+  kStringPiece,
+};
+
+StringType GetStringType(const FieldDescriptor& field) {
+  ABSL_CHECK_EQ(field.cpp_type(), FieldDescriptor::CPPTYPE_STRING);
+
+  if (field.options().has_ctype()) {
+    switch (field.options().ctype()) {
+      case FieldOptions::CORD:
+        return StringType::kCord;
+      case FieldOptions::STRING_PIECE:
+        return StringType::kStringPiece;
+      default:
+        return StringType::kString;
+    }
+  }
+
+  const pb::CppFeatures& cpp_features =
+      CppGenerator::GetResolvedSourceFeatures(field).GetExtension(::pb::cpp);
+  switch (cpp_features.string_type()) {
+    case pb::CppFeatures::CORD:
+      return StringType::kCord;
+    case pb::CppFeatures::VIEW:
+      return StringType::kView;
+    default:
+      return StringType::kString;
+  }
+}
+
 std::unique_ptr<FieldGeneratorBase> MakeGenerator(const FieldDescriptor* field,
                                                   const Options& options,
                                                   MessageSCCAnalyzer* scc) {
@@ -243,7 +279,7 @@ std::unique_ptr<FieldGeneratorBase> MakeGenerator(const FieldDescriptor* field,
       case FieldDescriptor::CPPTYPE_MESSAGE:
         return MakeRepeatedMessageGenerator(field, options, scc);
       case FieldDescriptor::CPPTYPE_STRING: {
-        if (field->cpp_string_type() == FieldDescriptor::CppStringType::kView) {
+        if (GetStringType(*field) == StringType::kView) {
           return MakeRepeatedStringViewGenerator(field, options, scc);
         } else {
           return MakeRepeatedStringGenerator(field, options, scc);
@@ -267,10 +303,10 @@ std::unique_ptr<FieldGeneratorBase> MakeGenerator(const FieldDescriptor* field,
     case FieldDescriptor::CPPTYPE_ENUM:
       return MakeSinguarEnumGenerator(field, options, scc);
     case FieldDescriptor::CPPTYPE_STRING: {
-      switch (field->cpp_string_type()) {
-        case FieldDescriptor::CppStringType::kView:
+      switch (GetStringType(*field)) {
+        case StringType::kView:
           return MakeSingularStringViewGenerator(field, options, scc);
-        case FieldDescriptor::CppStringType::kCord:
+        case StringType::kCord:
           if (field->type() == FieldDescriptor::TYPE_BYTES) {
             if (field->real_containing_oneof()) {
               return MakeOneofCordGenerator(field, options, scc);
