@@ -69,14 +69,13 @@ TEST_F(CppGeneratorTest, LegacyClosedEnumOnNonEnumField) {
                  R"schema(
     edition = "2023";
     import "google/protobuf/cpp_features.proto";
-    
+
     message Foo {
       int32 bar = 1 [features.(pb.cpp).legacy_closed_enum = true];
     })schema");
 
   RunProtoc(
-      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir "
-      "--experimental_editions foo.proto");
+      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir foo.proto");
 
   ExpectErrorSubstring(
       "Field Foo.bar specifies the legacy_closed_enum feature but has non-enum "
@@ -97,10 +96,11 @@ TEST_F(CppGeneratorTest, LegacyClosedEnum) {
     })schema");
 
   RunProtoc(
-      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir "
-      "--experimental_editions foo.proto");
+      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir foo.proto");
 
-  ExpectNoErrors();
+  ExpectWarningSubstring(
+      "foo.proto:9:16: warning: Feature pb.CppFeatures.legacy_closed_enum has "
+      "been deprecated in edition 2023");
 }
 
 TEST_F(CppGeneratorTest, LegacyClosedEnumInherited) {
@@ -109,7 +109,7 @@ TEST_F(CppGeneratorTest, LegacyClosedEnumInherited) {
     edition = "2023";
     import "google/protobuf/cpp_features.proto";
     option features.(pb.cpp).legacy_closed_enum = true;
-    
+
     enum TestEnum {
       TEST_ENUM_UNKNOWN = 0;
     }
@@ -119,10 +119,11 @@ TEST_F(CppGeneratorTest, LegacyClosedEnumInherited) {
     })schema");
 
   RunProtoc(
-      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir "
-      "--experimental_editions foo.proto");
+      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir foo.proto");
 
-  ExpectNoErrors();
+  ExpectWarningSubstring(
+      "foo.proto: warning: Feature pb.CppFeatures.legacy_closed_enum has "
+      "been deprecated in edition 2023");
 }
 
 TEST_F(CppGeneratorTest, LegacyClosedEnumImplicit) {
@@ -130,7 +131,7 @@ TEST_F(CppGeneratorTest, LegacyClosedEnumImplicit) {
     edition = "2023";
     import "google/protobuf/cpp_features.proto";
     option features.(pb.cpp).legacy_closed_enum = true;
-    
+
     enum TestEnum {
       TEST_ENUM_UNKNOWN = 0;
     }
@@ -140,12 +141,122 @@ TEST_F(CppGeneratorTest, LegacyClosedEnumImplicit) {
     }
   )schema");
 
+  RunProtoc(
+      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir foo.proto");
+
+  ExpectErrorSubstring(
+      "Field Foo.bar has a closed enum type with implicit presence.");
+}
+
+TEST_F(CppGeneratorTest, AllowStringTypeForEdition2023) {
+  CreateTempFile("foo.proto", R"schema(
+    edition = "2023";
+    import "google/protobuf/cpp_features.proto";
+
+    message Foo {
+      int32 bar = 1;
+      bytes baz = 2 [features.(pb.cpp).string_type = CORD];
+    }
+  )schema");
+
+  RunProtoc(
+      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir foo.proto");
+
+  ExpectNoErrors();
+}
+
+TEST_F(CppGeneratorTest, ErrorsOnBothStringTypeAndCtype) {
+  CreateTempFile("foo.proto", R"schema(
+    edition = "2023";
+    import "google/protobuf/cpp_features.proto";
+
+    message Foo {
+      int32 bar = 1;
+      bytes baz = 2 [ctype = CORD, features.(pb.cpp).string_type = VIEW];
+    }
+  )schema");
+
+  RunProtoc(
+      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir foo.proto");
+
+  ExpectErrorSubstring(
+      "Foo.baz specifies both string_type and ctype which is not supported.");
+}
+
+TEST_F(CppGeneratorTest, StringTypeForCord) {
+  CreateTempFile("foo.proto", R"schema(
+    edition = "2024";
+    import "google/protobuf/cpp_features.proto";
+
+    message Foo {
+      int32 bar = 1;
+      bytes baz = 2 [features.(pb.cpp).string_type = CORD];
+    }
+  )schema");
+
+  RunProtoc(
+      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir "
+      "--experimental_editions foo.proto");
+
+  ExpectNoErrors();
+}
+
+TEST_F(CppGeneratorTest, CtypeForCord) {
+  CreateTempFile("foo.proto", R"schema(
+    edition = "2023";
+
+    message Foo {
+      int32 bar = 1;
+      bytes baz = 2 [ctype = CORD];
+    }
+  )schema");
+
+  RunProtoc(
+      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir foo.proto");
+
+  ExpectNoErrors();
+}
+
+TEST_F(CppGeneratorTest, StringTypeForStringFieldsOnly) {
+  CreateTempFile("foo.proto", R"schema(
+    edition = "2024";
+    import "google/protobuf/cpp_features.proto";
+
+    message Foo {
+      int32 bar = 1;
+      int32 baz = 2 [features.(pb.cpp).string_type = CORD];
+    }
+  )schema");
+
   RunProtoc(
       "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir "
       "--experimental_editions foo.proto");
 
   ExpectErrorSubstring(
-      "Field Foo.bar has a closed enum type with implicit presence.");
+      "Field Foo.baz specifies string_type, but is not a string nor bytes "
+      "field.");
+}
+
+TEST_F(CppGeneratorTest, StringTypeCordNotForExtension) {
+  CreateTempFile("foo.proto", R"schema(
+    edition = "2024";
+    import "google/protobuf/cpp_features.proto";
+
+    message Foo {
+      extensions 1 to max;
+    }
+    extend Foo {
+      bytes bar = 1 [features.(pb.cpp).string_type = CORD];
+    }
+  )schema");
+
+  RunProtoc(
+      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir "
+      "--experimental_editions foo.proto");
+
+  ExpectErrorSubstring(
+      "Extension bar specifies string_type=CORD which is not supported for "
+      "extensions.");
 }
 
 TEST_F(CppGeneratorTest, CtypeOnNoneStringFieldTest) {
@@ -156,8 +267,7 @@ TEST_F(CppGeneratorTest, CtypeOnNoneStringFieldTest) {
       int32 bar = 1 [ctype=STRING];
     })schema");
   RunProtoc(
-      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir "
-      "--experimental_editions foo.proto");
+      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir foo.proto");
   ExpectErrorSubstring(
       "Field Foo.bar specifies ctype, but is not "
       "a string nor bytes field.");
@@ -174,8 +284,7 @@ TEST_F(CppGeneratorTest, CtypeOnExtensionTest) {
       bytes bar = 1 [ctype=CORD];
     })schema");
   RunProtoc(
-      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir "
-      "--experimental_editions foo.proto");
+      "protocol_compiler --proto_path=$tmpdir --cpp_out=$tmpdir foo.proto");
   ExpectErrorSubstring(
       "Extension bar specifies ctype=CORD which is "
       "not supported for extensions.");
