@@ -41,6 +41,7 @@
 #include "google/protobuf/unittest.pb.h"
 #include "google/protobuf/unittest_mset.pb.h"
 #include "google/protobuf/unittest_mset_wire_format.pb.h"
+#include "google/protobuf/unittest_proto3.pb.h"
 
 // Must be included last.
 #include "google/protobuf/port_def.inc"
@@ -69,6 +70,11 @@ class GeneratedMessageReflectionTestHelper {
                                         const FieldDescriptor* field) {
     return msg.GetReflection()->IsLazilyVerifiedLazyField(field);
   }
+  template <typename T>
+  static const T& GetRaw(const Message& msg, const FieldDescriptor* field) {
+    const Reflection* reflection = msg.GetReflection();
+    return reflection->GetRaw<T>(msg, field);
+  }
 };
 
 namespace {
@@ -162,6 +168,99 @@ TEST(GeneratedMessageReflectionTest, GetStringReferenceCopy) {
           &cord_scratch));
 }
 
+TEST(GeneratedMessageReflectionTest, GetStringView) {
+  unittest::TestAllTypes message;
+  TestUtil::SetAllFields(&message);
+
+  const Reflection* reflection = message.GetReflection();
+  Reflection::ScratchSpace scratch;
+
+  EXPECT_EQ("115",
+            reflection->GetStringView(message, F("optional_string"), scratch));
+  EXPECT_EQ("124", reflection->GetStringView(
+                       message, F("optional_string_piece"), scratch));
+  EXPECT_EQ("125",
+            reflection->GetStringView(message, F("optional_cord"), scratch));
+}
+
+TEST(GeneratedMessageReflectionTest, GetStringViewWithExtensions) {
+  unittest::TestAllExtensions message;
+  google::protobuf::FileDescriptor const* descriptor_file =
+      message.GetDescriptor()->file();
+  google::protobuf::FieldDescriptor const* string_ext =
+      descriptor_file->FindExtensionByName("optional_string_extension");
+  google::protobuf::FieldDescriptor const* string_piece_ext =
+      descriptor_file->FindExtensionByName("optional_string_piece_extension");
+  google::protobuf::FieldDescriptor const* cord_ext =
+      descriptor_file->FindExtensionByName("optional_cord_extension");
+  message.SetExtension(protobuf_unittest::optional_string_extension, "foo");
+  message.SetExtension(protobuf_unittest::optional_string_piece_extension, "bar");
+  message.SetExtension(protobuf_unittest::optional_cord_extension, "baz");
+  const Reflection* reflection = message.GetReflection();
+  Reflection::ScratchSpace scratch;
+
+  EXPECT_EQ("foo", reflection->GetStringView(message, string_ext, scratch));
+  EXPECT_EQ("bar",
+            reflection->GetStringView(message, string_piece_ext, scratch));
+  EXPECT_EQ("baz", reflection->GetStringView(message, cord_ext, scratch));
+}
+
+TEST(GeneratedMessageReflectionTest, GetStringViewWithOneof) {
+  unittest::TestOneof2 message;
+  const Reflection* reflection = message.GetReflection();
+  const FieldDescriptor* string_field =
+      message.GetDescriptor()->FindFieldByName("foo_string");
+  const FieldDescriptor* string_piece_field =
+      message.GetDescriptor()->FindFieldByName("foo_string_piece");
+  Reflection::ScratchSpace scratch;
+
+  message.set_foo_string("foo");
+  EXPECT_EQ("foo", reflection->GetStringView(message, string_field, scratch));
+  EXPECT_EQ("",
+            reflection->GetStringView(message, string_piece_field, scratch));
+
+}
+
+TEST(GeneratedMessageReflectionTest, GetRepeatedStringView) {
+  unittest::TestAllTypes message;
+  TestUtil::AddRepeatedFields1(&message);
+  TestUtil::AddRepeatedFields2(&message);
+
+  const Reflection* reflection = message.GetReflection();
+  Reflection::ScratchSpace scratch;
+
+  EXPECT_EQ("215", reflection->GetRepeatedStringView(
+                       message, F("repeated_string"), 0, scratch));
+  EXPECT_EQ("224", reflection->GetRepeatedStringView(
+                       message, F("repeated_string_piece"), 0, scratch));
+  EXPECT_EQ("225", reflection->GetRepeatedStringView(
+                       message, F("repeated_cord"), 0, scratch));
+}
+
+TEST(GeneratedMessageReflectionTest, GetRepeatedStringViewWithExtensions) {
+  unittest::TestAllExtensions message;
+  google::protobuf::FileDescriptor const* descriptor_file =
+      message.GetDescriptor()->file();
+  google::protobuf::FieldDescriptor const* string_ext =
+      descriptor_file->FindExtensionByName("repeated_string_extension");
+  google::protobuf::FieldDescriptor const* string_piece_ext =
+      descriptor_file->FindExtensionByName("repeated_string_piece_extension");
+  google::protobuf::FieldDescriptor const* cord_ext =
+      descriptor_file->FindExtensionByName("repeated_cord_extension");
+  message.AddExtension(protobuf_unittest::repeated_string_extension, "foo");
+  message.AddExtension(protobuf_unittest::repeated_string_piece_extension, "bar");
+  message.AddExtension(protobuf_unittest::repeated_cord_extension, "baz");
+  const Reflection* reflection = message.GetReflection();
+  Reflection::ScratchSpace scratch;
+
+  EXPECT_EQ("foo",
+            reflection->GetRepeatedStringView(message, string_ext, 0, scratch));
+  EXPECT_EQ("bar", reflection->GetRepeatedStringView(message, string_piece_ext,
+                                                     0, scratch));
+  EXPECT_EQ("baz",
+            reflection->GetRepeatedStringView(message, cord_ext, 0, scratch));
+}
+
 
 class GeneratedMessageReflectionSwapTest : public testing::TestWithParam<bool> {
  protected:
@@ -296,9 +395,9 @@ TEST_P(GeneratedMessageReflectionSwapTest, OneofBothSet) {
 
 TEST_P(GeneratedMessageReflectionSwapTest, SwapFields) {
   std::unique_ptr<unittest::TestAllTypes> lhs(
-      Arena::CreateMessage<unittest::TestAllTypes>(nullptr));
+      Arena::Create<unittest::TestAllTypes>(nullptr));
   std::unique_ptr<unittest::TestAllTypes> rhs(
-      Arena::CreateMessage<unittest::TestAllTypes>(nullptr));
+      Arena::Create<unittest::TestAllTypes>(nullptr));
   lhs->set_optional_double(12.3);
   lhs->mutable_repeated_int32()->Add(10);
   lhs->mutable_repeated_int32()->Add(20);
@@ -334,9 +433,9 @@ TEST_P(GeneratedMessageReflectionSwapTest, SwapFields) {
 
 TEST_P(GeneratedMessageReflectionSwapTest, SwapFieldsAll) {
   std::unique_ptr<unittest::TestAllTypes> lhs(
-      Arena::CreateMessage<unittest::TestAllTypes>(nullptr));
+      Arena::Create<unittest::TestAllTypes>(nullptr));
   std::unique_ptr<unittest::TestAllTypes> rhs(
-      Arena::CreateMessage<unittest::TestAllTypes>(nullptr));
+      Arena::Create<unittest::TestAllTypes>(nullptr));
 
   TestUtil::SetAllFields(rhs.get());
 
@@ -351,8 +450,8 @@ TEST_P(GeneratedMessageReflectionSwapTest, SwapFieldsAll) {
 
 TEST(GeneratedMessageReflectionTest, SwapFieldsAllOnDifferentArena) {
   Arena arena1, arena2;
-  auto* message1 = Arena::CreateMessage<unittest::TestAllTypes>(&arena1);
-  auto* message2 = Arena::CreateMessage<unittest::TestAllTypes>(&arena2);
+  auto* message1 = Arena::Create<unittest::TestAllTypes>(&arena1);
+  auto* message2 = Arena::Create<unittest::TestAllTypes>(&arena2);
 
   TestUtil::SetAllFields(message2);
 
@@ -367,9 +466,9 @@ TEST(GeneratedMessageReflectionTest, SwapFieldsAllOnDifferentArena) {
 
 TEST(GeneratedMessageReflectionTest, SwapFieldsAllOnArenaHeap) {
   Arena arena;
-  auto* message1 = Arena::CreateMessage<unittest::TestAllTypes>(&arena);
+  auto* message1 = Arena::Create<unittest::TestAllTypes>(&arena);
   std::unique_ptr<unittest::TestAllTypes> message2(
-      Arena::CreateMessage<unittest::TestAllTypes>(nullptr));
+      Arena::Create<unittest::TestAllTypes>(nullptr));
 
   TestUtil::SetAllFields(message2.get());
 
@@ -401,8 +500,8 @@ TEST(GeneratedMessageReflectionTest, SwapFieldsAllExtensionArenaHeap) {
   Arena arena;
 
   std::unique_ptr<unittest::TestAllExtensions> message1(
-      Arena::CreateMessage<unittest::TestAllExtensions>(nullptr));
-  auto* message2 = Arena::CreateMessage<unittest::TestAllExtensions>(&arena);
+      Arena::Create<unittest::TestAllExtensions>(nullptr));
+  auto* message2 = Arena::Create<unittest::TestAllExtensions>(&arena);
 
   TestUtil::SetAllExtensions(message1.get());
 
@@ -417,8 +516,8 @@ TEST(GeneratedMessageReflectionTest, SwapFieldsAllExtensionArenaHeap) {
 
 TEST(GeneratedMessageReflectionTest, UnsafeShallowSwapFieldsAll) {
   Arena arena;
-  auto* message1 = Arena::CreateMessage<unittest::TestAllTypes>(&arena);
-  auto* message2 = Arena::CreateMessage<unittest::TestAllTypes>(&arena);
+  auto* message1 = Arena::Create<unittest::TestAllTypes>(&arena);
+  auto* message2 = Arena::Create<unittest::TestAllTypes>(&arena);
 
   TestUtil::SetAllFields(message2);
 
@@ -452,8 +551,8 @@ TEST(GeneratedMessageReflectionTest, UnsafeShallowSwapFieldsAll) {
 
 TEST(GeneratedMessageReflectionTest, UnsafeShallowSwapFieldsMap) {
   Arena arena;
-  auto* message1 = Arena::CreateMessage<unittest::TestMap>(&arena);
-  auto* message2 = Arena::CreateMessage<unittest::TestMap>(&arena);
+  auto* message1 = Arena::Create<unittest::TestMap>(&arena);
+  auto* message2 = Arena::Create<unittest::TestMap>(&arena);
 
   MapTestUtil::SetMapFields(message2);
 
@@ -476,8 +575,8 @@ TEST(GeneratedMessageReflectionTest, UnsafeShallowSwapFieldsMap) {
 
 TEST(GeneratedMessageReflectionTest, UnsafeShallowSwapFieldsAllExtension) {
   Arena arena;
-  auto* message1 = Arena::CreateMessage<unittest::TestAllExtensions>(&arena);
-  auto* message2 = Arena::CreateMessage<unittest::TestAllExtensions>(&arena);
+  auto* message1 = Arena::Create<unittest::TestAllExtensions>(&arena);
+  auto* message2 = Arena::Create<unittest::TestAllExtensions>(&arena);
 
   TestUtil::SetAllExtensions(message1);
 
@@ -534,8 +633,8 @@ TEST(GeneratedMessageReflectionTest, SwapFieldsOneof) {
 
 TEST(GeneratedMessageReflectionTest, UnsafeShallowSwapFieldsOneof) {
   Arena arena;
-  auto* message1 = Arena::CreateMessage<unittest::TestOneof2>(&arena);
-  auto* message2 = Arena::CreateMessage<unittest::TestOneof2>(&arena);
+  auto* message1 = Arena::Create<unittest::TestOneof2>(&arena);
+  auto* message2 = Arena::Create<unittest::TestOneof2>(&arena);
   TestUtil::SetOneof1(message1);
 
   std::vector<const FieldDescriptor*> fields;
@@ -553,8 +652,8 @@ TEST(GeneratedMessageReflectionTest, UnsafeShallowSwapFieldsOneof) {
 TEST(GeneratedMessageReflectionTest,
      UnsafeShallowSwapFieldsOneofExpectShallow) {
   Arena arena;
-  auto* message1 = Arena::CreateMessage<unittest::TestOneof2>(&arena);
-  auto* message2 = Arena::CreateMessage<unittest::TestOneof2>(&arena);
+  auto* message1 = Arena::Create<unittest::TestOneof2>(&arena);
+  auto* message2 = Arena::Create<unittest::TestOneof2>(&arena);
   TestUtil::SetOneof1(message1);
   message1->mutable_foo_message()->set_moo_int(1000);
   auto* kept_foo_ptr = message1->mutable_foo_message();
@@ -789,7 +888,7 @@ TEST(GeneratedMessageReflectionTest, SetAllocatedMessageOnArenaTest) {
   unittest::TestAllTypes from_message2;
   Arena arena;
   unittest::TestAllTypes* to_message =
-      Arena::CreateMessage<unittest::TestAllTypes>(&arena);
+      Arena::Create<unittest::TestAllTypes>(&arena);
   TestUtil::ReflectionTester reflection_tester(
       unittest::TestAllTypes::descriptor());
   reflection_tester.SetAllFieldsViaReflection(&from_message1);
@@ -855,7 +954,7 @@ TEST(GeneratedMessageReflectionTest, SetAllocatedExtensionMessageTest) {
 TEST(GeneratedMessageReflectionTest, SetAllocatedExtensionMessageOnArenaTest) {
   Arena arena;
   unittest::TestAllExtensions* to_message =
-      Arena::CreateMessage<unittest::TestAllExtensions>(&arena);
+      Arena::Create<unittest::TestAllExtensions>(&arena);
   unittest::TestAllExtensions from_message1;
   unittest::TestAllExtensions from_message2;
   TestUtil::ReflectionTester reflection_tester(
@@ -1087,7 +1186,7 @@ TEST(GeneratedMessageReflectionTest, SetAllocatedOneofMessageOnArenaTest) {
   unittest::TestOneof2 from_message2;
   Arena arena;
   unittest::TestOneof2* to_message =
-      Arena::CreateMessage<unittest::TestOneof2>(&arena);
+      Arena::Create<unittest::TestOneof2>(&arena);
   const Descriptor* descriptor = unittest::TestOneof2::descriptor();
   const Reflection* reflection = to_message->GetReflection();
 
@@ -1212,7 +1311,7 @@ TEST(GeneratedMessageReflectionTest, ReleaseOneofMessageTest) {
 TEST(GeneratedMessageReflectionTest, ArenaReleaseMessageTest) {
   Arena arena;
   unittest::TestAllTypes* message =
-      Arena::CreateMessage<unittest::TestAllTypes>(&arena);
+      Arena::Create<unittest::TestAllTypes>(&arena);
   TestUtil::ReflectionTester reflection_tester(
       unittest::TestAllTypes::descriptor());
 
@@ -1236,7 +1335,7 @@ TEST(GeneratedMessageReflectionTest, ArenaReleaseMessageTest) {
 TEST(GeneratedMessageReflectionTest, ArenaReleaseExtensionMessageTest) {
   Arena arena;
   unittest::TestAllExtensions* message =
-      Arena::CreateMessage<unittest::TestAllExtensions>(&arena);
+      Arena::Create<unittest::TestAllExtensions>(&arena);
   TestUtil::ReflectionTester reflection_tester(
       unittest::TestAllExtensions::descriptor());
 
@@ -1259,8 +1358,7 @@ TEST(GeneratedMessageReflectionTest, ArenaReleaseExtensionMessageTest) {
 
 TEST(GeneratedMessageReflectionTest, ArenaReleaseOneofMessageTest) {
   Arena arena;
-  unittest::TestOneof2* message =
-      Arena::CreateMessage<unittest::TestOneof2>(&arena);
+  unittest::TestOneof2* message = Arena::Create<unittest::TestOneof2>(&arena);
   TestUtil::ReflectionTester::SetOneofViaReflection(message);
 
   const Descriptor* descriptor = unittest::TestOneof2::descriptor();
@@ -1659,6 +1757,15 @@ TEST(GeneratedMessageReflection, ListFieldsSorted) {
                           Pointee(Property(&FieldDescriptor::number, 101))));
 }
 
+TEST(GeneratedMessageReflection, SwapImplicitPresenceShouldWork) {
+  proto3_unittest::TestHasbits lhs, rhs;
+  rhs.mutable_child()->set_optional_int32(-1);
+  lhs.GetReflection()->Swap(&lhs, &rhs);
+  EXPECT_EQ(lhs.child().optional_int32(), -1);
+}
+
 }  // namespace
 }  // namespace protobuf
 }  // namespace google
+
+#include "google/protobuf/port_undef.inc"
