@@ -1,32 +1,9 @@
 // Protocol Buffers - Google's data interchange format
 // Copyright 2008 Google Inc.  All rights reserved.
-// https://developers.google.com/protocol-buffers/
 //
-// Redistribution and use in source and binary forms, with or without
-// modification, are permitted provided that the following conditions are
-// met:
-//
-//     * Redistributions of source code must retain the above copyright
-// notice, this list of conditions and the following disclaimer.
-//     * Redistributions in binary form must reproduce the above
-// copyright notice, this list of conditions and the following disclaimer
-// in the documentation and/or other materials provided with the
-// distribution.
-//     * Neither the name of Google Inc. nor the names of its
-// contributors may be used to endorse or promote products derived from
-// this software without specific prior written permission.
-//
-// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+// Use of this source code is governed by a BSD-style
+// license that can be found in the LICENSE file or at
+// https://developers.google.com/open-source/licenses/bsd
 
 // Author: kenton@google.com (Kenton Varda)
 //  Based on original Protocol Buffers design by
@@ -34,35 +11,37 @@
 //
 // This file contains tests and benchmarks.
 
-#include <google/protobuf/io/coded_stream.h>
+#include "google/protobuf/io/coded_stream.h"
 
 #include <limits.h>
 
+#include <algorithm>
 #include <memory>
+#include <string>
 #include <vector>
 
-#include <google/protobuf/stubs/common.h>
-#include <google/protobuf/stubs/logging.h>
-#include <google/protobuf/stubs/logging.h>
-#include <google/protobuf/io/zero_copy_stream_impl.h>
-#include <google/protobuf/testing/googletest.h>
+#include "google/protobuf/stubs/common.h"
 #include <gtest/gtest.h>
-#include <google/protobuf/stubs/casts.h>
+#include "absl/base/casts.h"
+#include "absl/base/log_severity.h"
+#include "absl/log/absl_check.h"
+#include "absl/log/absl_log.h"
+#include "absl/log/scoped_mock_log.h"
+#include "absl/strings/cord.h"
+#include "absl/strings/string_view.h"
+#include "google/protobuf/io/zero_copy_stream_impl.h"
 
-#include <google/protobuf/port_def.inc>
+#include "google/protobuf/testing/googletest.h"
 
-// This declares an unsigned long long integer literal in a portable way.
-// (The original macro is way too big and ruins my formatting.)
-#undef ULL
-#define ULL(x) PROTOBUF_ULONGLONG(x)
 
+// Must be included last.
+#include "google/protobuf/port_def.inc"
 
 namespace google {
 namespace protobuf {
 namespace io {
 namespace {
 
-
 // ===================================================================
 // Data-Driven Test Infrastructure
 
@@ -71,7 +50,7 @@ namespace {
 // run multiple times, once for each item in some input array.  TEST_1D
 // tests all cases in a single input array.  TEST_2D tests all
 // combinations of cases from two arrays.  The arrays must be statically
-// defined such that the GOOGLE_ARRAYSIZE() macro works on them.  Example:
+// defined such that the ABSL_ARRAYSIZE() macro works on them.  Example:
 //
 // int kCases[] = {1, 2, 3, 4}
 // TEST_1D(MyFixture, MyTest, kCases) {
@@ -83,7 +62,7 @@ namespace {
 // which failed will be printed.  The case type must be printable using
 // ostream::operator<<.
 
-// TODO(kenton):  gTest now supports "parameterized tests" which would be
+// TODO:  gTest now supports "parameterized tests" which would be
 //   a better way to accomplish this.  Rewrite when time permits.
 
 #define TEST_1D(FIXTURE, NAME, CASES)                             \
@@ -94,7 +73,7 @@ namespace {
   };                                                              \
                                                                   \
   TEST_F(FIXTURE##_##NAME##_DD, NAME) {                           \
-    for (int i = 0; i < GOOGLE_ARRAYSIZE(CASES); i++) {                  \
+    for (size_t i = 0; i < ABSL_ARRAYSIZE(CASES); i++) {          \
       SCOPED_TRACE(testing::Message()                             \
                    << #CASES " case #" << i << ": " << CASES[i]); \
       DoSingleCase(CASES[i]);                                     \
@@ -113,8 +92,8 @@ namespace {
   };                                                                        \
                                                                             \
   TEST_F(FIXTURE##_##NAME##_DD, NAME) {                                     \
-    for (int i = 0; i < GOOGLE_ARRAYSIZE(CASES1); i++) {                           \
-      for (int j = 0; j < GOOGLE_ARRAYSIZE(CASES2); j++) {                         \
+    for (size_t i = 0; i < ABSL_ARRAYSIZE(CASES1); i++) {                   \
+      for (size_t j = 0; j < ABSL_ARRAYSIZE(CASES2); j++) {                 \
         SCOPED_TRACE(testing::Message()                                     \
                      << #CASES1 " case #" << i << ": " << CASES1[i] << ", " \
                      << #CASES2 " case #" << j << ": " << CASES2[j]);       \
@@ -133,10 +112,10 @@ class CodedStreamTest : public testing::Test {
  protected:
   // Buffer used during most of the tests. This assumes tests run sequentially.
   static constexpr int kBufferSize = 1024 * 64;
-  static uint8 buffer_[kBufferSize];
+  static uint8_t buffer_[kBufferSize];
 };
 
-uint8 CodedStreamTest::buffer_[CodedStreamTest::kBufferSize];
+uint8_t CodedStreamTest::buffer_[CodedStreamTest::kBufferSize] = {};
 
 // We test each operation over a variety of block sizes to insure that
 // we test cases where reads or writes cross buffer boundaries, cases
@@ -145,14 +124,17 @@ uint8 CodedStreamTest::buffer_[CodedStreamTest::kBufferSize];
 // checks.
 const int kBlockSizes[] = {1, 2, 3, 5, 7, 13, 32, 1024};
 
+// In several ReadCord test functions, we either clear the Cord before ReadCord
+// calls or not.
+const bool kResetCords[] = {false, true};
 
 // -------------------------------------------------------------------
 // Varint tests.
 
 struct VarintCase {
-  uint8 bytes[10];  // Encoded bytes.
-  int size;         // Encoded size, in bytes.
-  uint64 value;     // Parsed value.
+  uint8_t bytes[10];  // Encoded bytes.
+  size_t size;        // Encoded size, in bytes.
+  uint64_t value;     // Parsed value.
 };
 
 inline std::ostream& operator<<(std::ostream& os, const VarintCase& c) {
@@ -168,24 +150,25 @@ VarintCase kVarintCases[] = {
     {{0xbe, 0xf7, 0x92, 0x84, 0x0b},
      5,  // 2961488830
      (0x3e << 0) | (0x77 << 7) | (0x12 << 14) | (0x04 << 21) |
-         (ULL(0x0b) << 28)},
+         (uint64_t{0x0bu} << 28)},
 
     // 64-bit
     {{0xbe, 0xf7, 0x92, 0x84, 0x1b},
      5,  // 7256456126
      (0x3e << 0) | (0x77 << 7) | (0x12 << 14) | (0x04 << 21) |
-         (ULL(0x1b) << 28)},
+         (uint64_t{0x1bu} << 28)},
     {{0x80, 0xe6, 0xeb, 0x9c, 0xc3, 0xc9, 0xa4, 0x49},
      8,  // 41256202580718336
      (0x00 << 0) | (0x66 << 7) | (0x6b << 14) | (0x1c << 21) |
-         (ULL(0x43) << 28) | (ULL(0x49) << 35) | (ULL(0x24) << 42) |
-         (ULL(0x49) << 49)},
+         (uint64_t{0x43u} << 28) | (uint64_t{0x49u} << 35) |
+         (uint64_t{0x24u} << 42) | (uint64_t{0x49u} << 49)},
     // 11964378330978735131
     {{0x9b, 0xa8, 0xf9, 0xc2, 0xbb, 0xd6, 0x80, 0x85, 0xa6, 0x01},
      10,
      (0x1b << 0) | (0x28 << 7) | (0x79 << 14) | (0x42 << 21) |
-         (ULL(0x3b) << 28) | (ULL(0x56) << 35) | (ULL(0x00) << 42) |
-         (ULL(0x05) << 49) | (ULL(0x26) << 56) | (ULL(0x01) << 63)},
+         (uint64_t{0x3bu} << 28) | (uint64_t{0x56u} << 35) |
+         (uint64_t{0x00u} << 42) | (uint64_t{0x05u} << 49) |
+         (uint64_t{0x26u} << 56) | (uint64_t{0x01u} << 63)},
 };
 
 TEST_2D(CodedStreamTest, ReadVarint32, kVarintCases, kBlockSizes) {
@@ -195,9 +178,9 @@ TEST_2D(CodedStreamTest, ReadVarint32, kVarintCases, kBlockSizes) {
   {
     CodedInputStream coded_input(&input);
 
-    uint32 value;
+    uint32_t value;
     EXPECT_TRUE(coded_input.ReadVarint32(&value));
-    EXPECT_EQ(static_cast<uint32>(kVarintCases_case.value), value);
+    EXPECT_EQ(static_cast<uint32_t>(kVarintCases_case.value), value);
   }
 
   EXPECT_EQ(kVarintCases_case.size, input.ByteCount());
@@ -210,7 +193,7 @@ TEST_2D(CodedStreamTest, ReadTag, kVarintCases, kBlockSizes) {
   {
     CodedInputStream coded_input(&input);
 
-    uint32 expected_value = static_cast<uint32>(kVarintCases_case.value);
+    uint32_t expected_value = static_cast<uint32_t>(kVarintCases_case.value);
     EXPECT_EQ(expected_value, coded_input.ReadTag());
 
     EXPECT_TRUE(coded_input.LastTagWas(expected_value));
@@ -229,13 +212,15 @@ TEST_F(CodedStreamTest, EmptyInputBeforeEos) {
 
    private:
     bool Next(const void** data, int* size) override {
-      *data = NULL;
+      *data = nullptr;
       *size = 0;
       return count_++ < 2;
     }
-    void BackUp(int count) override { GOOGLE_LOG(FATAL) << "Tests never call this."; }
+    void BackUp(int count) override {
+      ABSL_LOG(FATAL) << "Tests never call this.";
+    }
     bool Skip(int count) override {
-      GOOGLE_LOG(FATAL) << "Tests never call this.";
+      ABSL_LOG(FATAL) << "Tests never call this.";
       return false;
     }
     int64_t ByteCount() const override { return 0; }
@@ -258,11 +243,11 @@ TEST_1D(CodedStreamTest, ExpectTag, kVarintCases) {
 
     // Read one byte to force coded_input.Refill() to be called.  Otherwise,
     // ExpectTag() will return a false negative.
-    uint8 dummy;
+    uint8_t dummy;
     coded_input.ReadRaw(&dummy, 1);
     EXPECT_EQ((uint)'\0', (uint)dummy);
 
-    uint32 expected_value = static_cast<uint32>(kVarintCases_case.value);
+    uint32_t expected_value = static_cast<uint32_t>(kVarintCases_case.value);
 
     // ExpectTag() produces false negatives for large values.
     if (kVarintCases_case.size <= 2) {
@@ -283,16 +268,17 @@ TEST_1D(CodedStreamTest, ExpectTag, kVarintCases) {
 TEST_1D(CodedStreamTest, ExpectTagFromArray, kVarintCases) {
   memcpy(buffer_, kVarintCases_case.bytes, kVarintCases_case.size);
 
-  const uint32 expected_value = static_cast<uint32>(kVarintCases_case.value);
+  const uint32_t expected_value =
+      static_cast<uint32_t>(kVarintCases_case.value);
 
   // If the expectation succeeds, it should return a pointer past the tag.
   if (kVarintCases_case.size <= 2) {
-    EXPECT_TRUE(NULL == CodedInputStream::ExpectTagFromArray(
+    EXPECT_TRUE(nullptr == CodedInputStream::ExpectTagFromArray(
                             buffer_, expected_value + 1));
     EXPECT_TRUE(buffer_ + kVarintCases_case.size ==
                 CodedInputStream::ExpectTagFromArray(buffer_, expected_value));
   } else {
-    EXPECT_TRUE(NULL ==
+    EXPECT_TRUE(nullptr ==
                 CodedInputStream::ExpectTagFromArray(buffer_, expected_value));
   }
 }
@@ -304,7 +290,7 @@ TEST_2D(CodedStreamTest, ReadVarint64, kVarintCases, kBlockSizes) {
   {
     CodedInputStream coded_input(&input);
 
-    uint64 value;
+    uint64_t value;
     EXPECT_TRUE(coded_input.ReadVarint64(&value));
     EXPECT_EQ(kVarintCases_case.value, value);
   }
@@ -313,7 +299,7 @@ TEST_2D(CodedStreamTest, ReadVarint64, kVarintCases, kBlockSizes) {
 }
 
 TEST_2D(CodedStreamTest, WriteVarint32, kVarintCases, kBlockSizes) {
-  if (kVarintCases_case.value > ULL(0x00000000FFFFFFFF)) {
+  if (kVarintCases_case.value > uint64_t{0x00000000FFFFFFFFu}) {
     // Skip this test for the 64-bit values.
     return;
   }
@@ -323,7 +309,7 @@ TEST_2D(CodedStreamTest, WriteVarint32, kVarintCases, kBlockSizes) {
   {
     CodedOutputStream coded_output(&output);
 
-    coded_output.WriteVarint32(static_cast<uint32>(kVarintCases_case.value));
+    coded_output.WriteVarint32(static_cast<uint32_t>(kVarintCases_case.value));
     EXPECT_FALSE(coded_output.HadError());
 
     EXPECT_EQ(kVarintCases_case.size, coded_output.ByteCount());
@@ -355,7 +341,7 @@ TEST_2D(CodedStreamTest, WriteVarint64, kVarintCases, kBlockSizes) {
 //   "sorry, unimplemented: `method_call_expr' not supported by dump_expr"
 #if !defined(__GNUC__) || __GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ > 3)
 
-int32 kSignExtendedVarintCases[] = {0, 1, -1, 1237894, -37895138};
+int32_t kSignExtendedVarintCases[] = {0, 1, -1, 1237894, -37895138};
 
 TEST_2D(CodedStreamTest, WriteVarint32SignExtended, kSignExtendedVarintCases,
         kBlockSizes) {
@@ -386,10 +372,10 @@ TEST_2D(CodedStreamTest, WriteVarint32SignExtended, kSignExtendedVarintCases,
   {
     CodedInputStream coded_input(&input);
 
-    uint64 value;
+    uint64_t value;
     EXPECT_TRUE(coded_input.ReadVarint64(&value));
 
-    EXPECT_EQ(kSignExtendedVarintCases_case, static_cast<int64>(value));
+    EXPECT_EQ(kSignExtendedVarintCases_case, static_cast<int64_t>(value));
   }
 
   EXPECT_EQ(output.ByteCount(), input.ByteCount());
@@ -402,8 +388,8 @@ TEST_2D(CodedStreamTest, WriteVarint32SignExtended, kSignExtendedVarintCases,
 // Varint failure test.
 
 struct VarintErrorCase {
-  uint8 bytes[12];
-  int size;
+  uint8_t bytes[12];
+  size_t size;
   bool can_parse;
 };
 
@@ -433,22 +419,22 @@ const VarintErrorCase kVarintErrorCases[] = {
 
 TEST_2D(CodedStreamTest, ReadVarint32Error, kVarintErrorCases, kBlockSizes) {
   memcpy(buffer_, kVarintErrorCases_case.bytes, kVarintErrorCases_case.size);
-  ArrayInputStream input(buffer_, kVarintErrorCases_case.size,
+  ArrayInputStream input(buffer_, static_cast<int>(kVarintErrorCases_case.size),
                          kBlockSizes_case);
   CodedInputStream coded_input(&input);
 
-  uint32 value;
+  uint32_t value;
   EXPECT_EQ(kVarintErrorCases_case.can_parse, coded_input.ReadVarint32(&value));
 }
 
 TEST_2D(CodedStreamTest, ReadVarint32Error_LeavesValueInInitializedState,
         kVarintErrorCases, kBlockSizes) {
   memcpy(buffer_, kVarintErrorCases_case.bytes, kVarintErrorCases_case.size);
-  ArrayInputStream input(buffer_, kVarintErrorCases_case.size,
+  ArrayInputStream input(buffer_, static_cast<int>(kVarintErrorCases_case.size),
                          kBlockSizes_case);
   CodedInputStream coded_input(&input);
 
-  uint32 value = 0;
+  uint32_t value = 0;
   EXPECT_EQ(kVarintErrorCases_case.can_parse, coded_input.ReadVarint32(&value));
   // While the specific value following a failure is not critical, we do want to
   // ensure that it doesn't get set to an uninitialized value. (This check fails
@@ -458,22 +444,22 @@ TEST_2D(CodedStreamTest, ReadVarint32Error_LeavesValueInInitializedState,
 
 TEST_2D(CodedStreamTest, ReadVarint64Error, kVarintErrorCases, kBlockSizes) {
   memcpy(buffer_, kVarintErrorCases_case.bytes, kVarintErrorCases_case.size);
-  ArrayInputStream input(buffer_, kVarintErrorCases_case.size,
+  ArrayInputStream input(buffer_, static_cast<int>(kVarintErrorCases_case.size),
                          kBlockSizes_case);
   CodedInputStream coded_input(&input);
 
-  uint64 value;
+  uint64_t value;
   EXPECT_EQ(kVarintErrorCases_case.can_parse, coded_input.ReadVarint64(&value));
 }
 
 TEST_2D(CodedStreamTest, ReadVarint64Error_LeavesValueInInitializedState,
         kVarintErrorCases, kBlockSizes) {
   memcpy(buffer_, kVarintErrorCases_case.bytes, kVarintErrorCases_case.size);
-  ArrayInputStream input(buffer_, kVarintErrorCases_case.size,
+  ArrayInputStream input(buffer_, static_cast<int>(kVarintErrorCases_case.size),
                          kBlockSizes_case);
   CodedInputStream coded_input(&input);
 
-  uint64 value = 0;
+  uint64_t value = 0;
   EXPECT_EQ(kVarintErrorCases_case.can_parse, coded_input.ReadVarint64(&value));
   // While the specific value following a failure is not critical, we do want to
   // ensure that it doesn't get set to an uninitialized value. (This check fails
@@ -485,7 +471,7 @@ TEST_2D(CodedStreamTest, ReadVarint64Error_LeavesValueInInitializedState,
 // VarintSize
 
 struct VarintSizeCase {
-  uint64 value;
+  uint64_t value;
   int size;
 };
 
@@ -500,8 +486,8 @@ VarintSizeCase kVarintSizeCases[] = {
     {128u, 2},
     {758923u, 3},
     {4000000000u, 5},
-    {ULL(41256202580718336), 8},
-    {ULL(11964378330978735131), 10},
+    {uint64_t{41256202580718336u}, 8},
+    {uint64_t{11964378330978735131u}, 10},
 };
 
 TEST_1D(CodedStreamTest, VarintSize32, kVarintSizeCases) {
@@ -512,7 +498,7 @@ TEST_1D(CodedStreamTest, VarintSize32, kVarintSizeCases) {
 
   EXPECT_EQ(kVarintSizeCases_case.size,
             CodedOutputStream::VarintSize32(
-                static_cast<uint32>(kVarintSizeCases_case.value)));
+                static_cast<uint32_t>(kVarintSizeCases_case.value)));
 }
 
 TEST_1D(CodedStreamTest, VarintSize64, kVarintSizeCases) {
@@ -526,8 +512,8 @@ TEST_F(CodedStreamTest, VarintSize32PowersOfTwo) {
     if (i % 7 == 0) {
       expected += 1;
     }
-    EXPECT_EQ(expected,
-              CodedOutputStream::VarintSize32(static_cast<uint32>(0x1u << i)));
+    EXPECT_EQ(expected, CodedOutputStream::VarintSize32(
+                            static_cast<uint32_t>(0x1u << i)));
   }
 }
 
@@ -537,8 +523,7 @@ TEST_F(CodedStreamTest, VarintSize64PowersOfTwo) {
     if (i % 7 == 0) {
       expected += 1;
     }
-    EXPECT_EQ(expected, CodedOutputStream::VarintSize64(
-                            static_cast<uint64>(0x1ull << i)));
+    EXPECT_EQ(expected, CodedOutputStream::VarintSize64(uint64_t{1} << i));
   }
 }
 
@@ -546,13 +531,13 @@ TEST_F(CodedStreamTest, VarintSize64PowersOfTwo) {
 // Fixed-size int tests
 
 struct Fixed32Case {
-  uint8 bytes[sizeof(uint32)];  // Encoded bytes.
-  uint32 value;                 // Parsed value.
+  uint8_t bytes[sizeof(uint32_t)];  // Encoded bytes.
+  uint32_t value;                   // Parsed value.
 };
 
 struct Fixed64Case {
-  uint8 bytes[sizeof(uint64)];  // Encoded bytes.
-  uint64 value;                 // Parsed value.
+  uint8_t bytes[sizeof(uint64_t)];  // Encoded bytes.
+  uint64_t value;                   // Parsed value.
 };
 
 inline std::ostream& operator<<(std::ostream& os, const Fixed32Case& c) {
@@ -569,8 +554,10 @@ Fixed32Case kFixed32Cases[] = {
 };
 
 Fixed64Case kFixed64Cases[] = {
-    {{0xef, 0xcd, 0xab, 0x90, 0x12, 0x34, 0x56, 0x78}, ULL(0x7856341290abcdef)},
-    {{0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88}, ULL(0x8877665544332211)},
+    {{0xef, 0xcd, 0xab, 0x90, 0x12, 0x34, 0x56, 0x78},
+     uint64_t{0x7856341290abcdefu}},
+    {{0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88},
+     uint64_t{0x8877665544332211u}},
 };
 
 TEST_2D(CodedStreamTest, ReadLittleEndian32, kFixed32Cases, kBlockSizes) {
@@ -580,12 +567,12 @@ TEST_2D(CodedStreamTest, ReadLittleEndian32, kFixed32Cases, kBlockSizes) {
   {
     CodedInputStream coded_input(&input);
 
-    uint32 value;
+    uint32_t value;
     EXPECT_TRUE(coded_input.ReadLittleEndian32(&value));
     EXPECT_EQ(kFixed32Cases_case.value, value);
   }
 
-  EXPECT_EQ(sizeof(uint32), input.ByteCount());
+  EXPECT_EQ(sizeof(uint32_t), input.ByteCount());
 }
 
 TEST_2D(CodedStreamTest, ReadLittleEndian64, kFixed64Cases, kBlockSizes) {
@@ -595,12 +582,12 @@ TEST_2D(CodedStreamTest, ReadLittleEndian64, kFixed64Cases, kBlockSizes) {
   {
     CodedInputStream coded_input(&input);
 
-    uint64 value;
+    uint64_t value;
     EXPECT_TRUE(coded_input.ReadLittleEndian64(&value));
     EXPECT_EQ(kFixed64Cases_case.value, value);
   }
 
-  EXPECT_EQ(sizeof(uint64), input.ByteCount());
+  EXPECT_EQ(sizeof(uint64_t), input.ByteCount());
 }
 
 TEST_2D(CodedStreamTest, WriteLittleEndian32, kFixed32Cases, kBlockSizes) {
@@ -612,11 +599,11 @@ TEST_2D(CodedStreamTest, WriteLittleEndian32, kFixed32Cases, kBlockSizes) {
     coded_output.WriteLittleEndian32(kFixed32Cases_case.value);
     EXPECT_FALSE(coded_output.HadError());
 
-    EXPECT_EQ(sizeof(uint32), coded_output.ByteCount());
+    EXPECT_EQ(sizeof(uint32_t), coded_output.ByteCount());
   }
 
-  EXPECT_EQ(sizeof(uint32), output.ByteCount());
-  EXPECT_EQ(0, memcmp(buffer_, kFixed32Cases_case.bytes, sizeof(uint32)));
+  EXPECT_EQ(sizeof(uint32_t), output.ByteCount());
+  EXPECT_EQ(0, memcmp(buffer_, kFixed32Cases_case.bytes, sizeof(uint32_t)));
 }
 
 TEST_2D(CodedStreamTest, WriteLittleEndian64, kFixed64Cases, kBlockSizes) {
@@ -628,11 +615,11 @@ TEST_2D(CodedStreamTest, WriteLittleEndian64, kFixed64Cases, kBlockSizes) {
     coded_output.WriteLittleEndian64(kFixed64Cases_case.value);
     EXPECT_FALSE(coded_output.HadError());
 
-    EXPECT_EQ(sizeof(uint64), coded_output.ByteCount());
+    EXPECT_EQ(sizeof(uint64_t), coded_output.ByteCount());
   }
 
-  EXPECT_EQ(sizeof(uint64), output.ByteCount());
-  EXPECT_EQ(0, memcmp(buffer_, kFixed64Cases_case.bytes, sizeof(uint64)));
+  EXPECT_EQ(sizeof(uint64_t), output.ByteCount());
+  EXPECT_EQ(0, memcmp(buffer_, kFixed64Cases_case.bytes, sizeof(uint64_t)));
 }
 
 // Tests using the static methods to read fixed-size values from raw arrays.
@@ -640,8 +627,8 @@ TEST_2D(CodedStreamTest, WriteLittleEndian64, kFixed64Cases, kBlockSizes) {
 TEST_1D(CodedStreamTest, ReadLittleEndian32FromArray, kFixed32Cases) {
   memcpy(buffer_, kFixed32Cases_case.bytes, sizeof(kFixed32Cases_case.bytes));
 
-  uint32 value;
-  const uint8* end =
+  uint32_t value;
+  const uint8_t* end =
       CodedInputStream::ReadLittleEndian32FromArray(buffer_, &value);
   EXPECT_EQ(kFixed32Cases_case.value, value);
   EXPECT_TRUE(end == buffer_ + sizeof(value));
@@ -650,8 +637,8 @@ TEST_1D(CodedStreamTest, ReadLittleEndian32FromArray, kFixed32Cases) {
 TEST_1D(CodedStreamTest, ReadLittleEndian64FromArray, kFixed64Cases) {
   memcpy(buffer_, kFixed64Cases_case.bytes, sizeof(kFixed64Cases_case.bytes));
 
-  uint64 value;
-  const uint8* end =
+  uint64_t value;
+  const uint8_t* end =
       CodedInputStream::ReadLittleEndian64FromArray(buffer_, &value);
   EXPECT_EQ(kFixed64Cases_case.value, value);
   EXPECT_TRUE(end == buffer_ + sizeof(value));
@@ -724,14 +711,14 @@ TEST_1D(CodedStreamTest, ReadStringImpossiblyLarge, kBlockSizes) {
 TEST_F(CodedStreamTest, ReadStringImpossiblyLargeFromStringOnStack) {
   // Same test as above, except directly use a buffer. This used to cause
   // crashes while the above did not.
-  uint8 buffer[8];
+  uint8_t buffer[8] = {};
   CodedInputStream coded_input(buffer, 8);
   std::string str;
   EXPECT_FALSE(coded_input.ReadString(&str, 1 << 30));
 }
 
 TEST_F(CodedStreamTest, ReadStringImpossiblyLargeFromStringOnHeap) {
-  std::unique_ptr<uint8[]> buffer(new uint8[8]);
+  std::unique_ptr<uint8_t[]> buffer(new uint8_t[8]);
   CodedInputStream coded_input(buffer.get(), 8);
   std::string str;
   EXPECT_FALSE(coded_input.ReadString(&str, 1 << 30));
@@ -743,7 +730,7 @@ TEST_1D(CodedStreamTest, ReadStringReservesMemoryOnTotalLimit, kBlockSizes) {
 
   {
     CodedInputStream coded_input(&input);
-    coded_input.SetTotalBytesLimit(sizeof(kRawBytes), sizeof(kRawBytes));
+    coded_input.SetTotalBytesLimit(sizeof(kRawBytes));
     EXPECT_EQ(sizeof(kRawBytes), coded_input.BytesUntilTotalBytesLimit());
 
     std::string str;
@@ -751,7 +738,7 @@ TEST_1D(CodedStreamTest, ReadStringReservesMemoryOnTotalLimit, kBlockSizes) {
     EXPECT_EQ(sizeof(kRawBytes) - strlen(kRawBytes),
               coded_input.BytesUntilTotalBytesLimit());
     EXPECT_EQ(kRawBytes, str);
-    // TODO(liujisi): Replace with a more meaningful test (see cl/60966023).
+    // TODO: Replace with a more meaningful test (see cl/60966023).
     EXPECT_GE(str.capacity(), strlen(kRawBytes));
   }
 
@@ -769,7 +756,7 @@ TEST_1D(CodedStreamTest, ReadStringReservesMemoryOnPushedLimit, kBlockSizes) {
     std::string str;
     EXPECT_TRUE(coded_input.ReadString(&str, strlen(kRawBytes)));
     EXPECT_EQ(kRawBytes, str);
-    // TODO(liujisi): Replace with a more meaningful test (see cl/60966023).
+    // TODO: Replace with a more meaningful test (see cl/60966023).
     EXPECT_GE(str.capacity(), strlen(kRawBytes));
   }
 
@@ -793,7 +780,7 @@ TEST_F(CodedStreamTest, ReadStringNoReservationIfLimitsNotSet) {
     // expects that string will allocate more than strlen(kRawBytes)
     // if the content of kRawBytes is appended to string in small
     // chunks.
-    // TODO(liujisi): Replace with a more meaningful test (see cl/60966023).
+    // TODO: Replace with a more meaningful test (see cl/60966023).
     EXPECT_GE(str.capacity(), strlen(kRawBytes));
   }
 
@@ -866,7 +853,7 @@ TEST_F(CodedStreamTest, ReadStringNoReservationSizeIsOverTheTotalBytesLimit) {
 
   {
     CodedInputStream coded_input(&input);
-    coded_input.SetTotalBytesLimit(16, 16);
+    coded_input.SetTotalBytesLimit(16);
 
     std::string str;
     EXPECT_FALSE(coded_input.ReadString(&str, strlen(kRawBytes)));
@@ -888,7 +875,7 @@ TEST_F(CodedStreamTest,
   {
     CodedInputStream coded_input(&input);
     coded_input.PushLimit(sizeof(buffer_));
-    coded_input.SetTotalBytesLimit(16, 16);
+    coded_input.SetTotalBytesLimit(16);
 
     std::string str;
     EXPECT_FALSE(coded_input.ReadString(&str, strlen(kRawBytes)));
@@ -910,7 +897,7 @@ TEST_F(CodedStreamTest,
   {
     CodedInputStream coded_input(&input);
     coded_input.PushLimit(16);
-    coded_input.SetTotalBytesLimit(sizeof(buffer_), sizeof(buffer_));
+    coded_input.SetTotalBytesLimit(sizeof(buffer_));
     EXPECT_EQ(sizeof(buffer_), coded_input.BytesUntilTotalBytesLimit());
 
     std::string str;
@@ -923,6 +910,186 @@ TEST_F(CodedStreamTest,
 }
 
 
+// -------------------------------------------------------------------
+// Cord reads and writes
+
+TEST_1D(CodedStreamTest, ReadCord, kBlockSizes) {
+  memcpy(buffer_, kRawBytes, sizeof(kRawBytes));
+  ArrayInputStream input(buffer_, sizeof(buffer_), kBlockSizes_case);
+
+  {
+    CodedInputStream coded_input(&input);
+
+    absl::Cord cord;
+    EXPECT_TRUE(coded_input.ReadCord(&cord, strlen(kRawBytes)));
+    EXPECT_EQ(absl::Cord(kRawBytes), cord);
+  }
+
+  EXPECT_EQ(strlen(kRawBytes), input.ByteCount());
+}
+
+TEST_1D(CodedStreamTest, ReadCordReuseCord, kBlockSizes) {
+  ASSERT_GT(sizeof(buffer_), 1362 * sizeof(kRawBytes));
+  for (size_t i = 0; i < 1362; i++) {
+    memcpy(buffer_ + i * sizeof(kRawBytes), kRawBytes, sizeof(kRawBytes));
+  }
+  ArrayInputStream input(buffer_, sizeof(buffer_), kBlockSizes_case);
+
+  int total_read = 0;
+  {
+    const char* buffer_str = reinterpret_cast<const char*>(buffer_);
+    CodedInputStream coded_input(&input);
+    static const int kSizes[] = {0,  1,  2,  3,   4,    5,    6,    7,
+                                 8,  9,  10, 11,  12,   13,   14,   15,
+                                 16, 17, 50, 100, 1023, 1024, 8000, 16000};
+    int total_size = 0;
+    std::vector<int> sizes;
+    for (size_t i = 0; i < ABSL_ARRAYSIZE(kSizes); ++i) {
+      sizes.push_back(kSizes[i]);
+      total_size += kSizes[i];
+    }
+    ASSERT_GE(1362 * sizeof(kRawBytes), total_size * 2);
+
+    absl::Cord reused_cord;
+    for (int pass = 0; pass < 2; ++pass) {
+      for (size_t i = 0; i < sizes.size(); ++i) {
+        const int size = sizes[i];
+        EXPECT_TRUE(coded_input.ReadCord(&reused_cord, size));
+        EXPECT_EQ(size, reused_cord.size());
+        EXPECT_EQ(
+            std::string(buffer_str + total_read, static_cast<size_t>(size)),
+            std::string(reused_cord));
+        total_read += size;
+      }
+      std::reverse(sizes.begin(), sizes.end());  // Second pass is in reverse.
+    }
+  }
+  EXPECT_EQ(total_read, input.ByteCount());
+}
+
+TEST_2D(CodedStreamTest, ReadCordWithLimit, kBlockSizes, kResetCords) {
+  memcpy(buffer_, kRawBytes, strlen(kRawBytes));
+  ArrayInputStream input(buffer_, strlen(kRawBytes), kBlockSizes_case);
+  CodedInputStream coded_input(&input);
+
+  CodedInputStream::Limit limit = coded_input.PushLimit(10);
+  absl::Cord cord;
+  EXPECT_TRUE(coded_input.ReadCord(&cord, 5));
+  EXPECT_EQ(5, coded_input.BytesUntilLimit());
+  if (kResetCords_case) cord.Clear();
+  EXPECT_TRUE(coded_input.ReadCord(&cord, 4));
+  EXPECT_EQ(1, coded_input.BytesUntilLimit());
+  if (kResetCords_case) cord.Clear();
+  EXPECT_FALSE(coded_input.ReadCord(&cord, 2));
+  EXPECT_EQ(0, coded_input.BytesUntilLimit());
+  EXPECT_EQ(1, cord.size());
+
+  coded_input.PopLimit(limit);
+
+  if (kResetCords_case) cord.Clear();
+  EXPECT_TRUE(coded_input.ReadCord(&cord, strlen(kRawBytes) - 10));
+  EXPECT_EQ(std::string(kRawBytes + 10), std::string(cord));
+}
+
+TEST_1D(CodedStreamTest, ReadLargeCord, kResetCords) {
+  absl::Cord large_cord;
+  for (int i = 0; i < 1024; i++) {
+    large_cord.Append(kRawBytes);
+  }
+  CordInputStream input(&large_cord);
+
+  {
+    CodedInputStream coded_input(&input);
+
+    absl::Cord cord;
+    if (!kResetCords_case) cord.Append(absl::string_view("value"));
+    EXPECT_TRUE(
+        coded_input.ReadCord(&cord, static_cast<int>(large_cord.size())));
+    EXPECT_EQ(large_cord, cord);
+  }
+
+  EXPECT_EQ(large_cord.size(), input.ByteCount());
+}
+
+// Check to make sure ReadString doesn't crash on impossibly large strings.
+TEST_2D(CodedStreamTest, ReadCordImpossiblyLarge, kBlockSizes, kResetCords) {
+  ArrayInputStream input(buffer_, sizeof(buffer_), kBlockSizes_case);
+
+  {
+    CodedInputStream coded_input(&input);
+
+    absl::Cord cord;
+    if (!kResetCords_case) cord.Append(absl::string_view("value"));
+    // Try to read a gigabyte.  This should fail because the input is only
+    // sizeof(buffer_) bytes.
+    EXPECT_FALSE(coded_input.ReadCord(&cord, 1 << 30));
+  }
+}
+
+TEST_1D(CodedStreamTest, WriteCord, kBlockSizes) {
+  ArrayOutputStream output(buffer_, sizeof(buffer_), kBlockSizes_case);
+
+  {
+    CodedOutputStream coded_output(&output);
+
+    absl::Cord cord(kRawBytes);
+    coded_output.WriteCord(cord);
+    EXPECT_FALSE(coded_output.HadError());
+
+    EXPECT_EQ(strlen(kRawBytes), coded_output.ByteCount());
+  }
+
+  EXPECT_EQ(strlen(kRawBytes), output.ByteCount());
+  EXPECT_EQ(0, memcmp(buffer_, kRawBytes, strlen(kRawBytes)));
+}
+
+TEST_F(CodedStreamTest, WriteLargeCord) {
+  absl::Cord large_cord;
+  for (int i = 0; i < 1024; i++) {
+    large_cord.Append(kRawBytes);
+  }
+
+  CordOutputStream output;
+  {
+    CodedOutputStream coded_output(&output);
+
+    coded_output.WriteCord(large_cord);
+    EXPECT_FALSE(coded_output.HadError());
+
+    EXPECT_EQ(large_cord.size(), coded_output.ByteCount());
+    EXPECT_EQ(large_cord.size(), output.ByteCount());
+  }
+  absl::Cord output_cord = output.Consume();
+  EXPECT_EQ(large_cord, output_cord);
+}
+
+TEST_F(CodedStreamTest, Trim) {
+  CordOutputStream cord_output;
+  CodedOutputStream coded_output(&cord_output);
+
+  // Verify that any initially reserved output buffers created when the output
+  // streams were created are trimmed on an initial Trim call.
+  coded_output.Trim();
+  EXPECT_EQ(0, coded_output.ByteCount());
+
+  // Write a single byte to the coded stream, ensure the cord stream has been
+  // advanced, and then verify Trim() does the right thing.
+  const char kTestData[] = "abcdef";
+  coded_output.WriteRaw(kTestData, 1);
+  coded_output.Trim();
+  EXPECT_EQ(1, coded_output.ByteCount());
+
+  // Write some more data to the coded stream, Trim() it, and verify
+  // everything behaves as expected.
+  coded_output.WriteRaw(kTestData, sizeof(kTestData));
+  coded_output.Trim();
+  EXPECT_EQ(1 + sizeof(kTestData), coded_output.ByteCount());
+
+  absl::Cord cord = cord_output.Consume();
+  EXPECT_EQ(1 + sizeof(kTestData), cord.size());
+}
+
+
 // -------------------------------------------------------------------
 // Skip
 
@@ -1022,7 +1189,7 @@ TEST_1D(CodedStreamTest, BasicLimit, kBlockSizes) {
     CodedInputStream::Limit limit = coded_input.PushLimit(8);
 
     // Read until we hit the limit.
-    uint32 value;
+    uint32_t value;
     EXPECT_EQ(8, coded_input.BytesUntilLimit());
     EXPECT_TRUE(coded_input.ReadLittleEndian32(&value));
     EXPECT_EQ(4, coded_input.BytesUntilLimit());
@@ -1053,7 +1220,7 @@ TEST_1D(CodedStreamTest, SmallLimitOnTopOfBigLimit, kBlockSizes) {
     EXPECT_EQ(8, coded_input.BytesUntilLimit());
     CodedInputStream::Limit limit2 = coded_input.PushLimit(4);
 
-    uint32 value;
+    uint32_t value;
 
     // Read until we hit limit2, the top and shortest limit.
     EXPECT_EQ(4, coded_input.BytesUntilLimit());
@@ -1095,7 +1262,7 @@ TEST_1D(CodedStreamTest, BigLimitOnTopOfSmallLimit, kBlockSizes) {
     EXPECT_EQ(4, coded_input.BytesUntilLimit());
     CodedInputStream::Limit limit2 = coded_input.PushLimit(8);
 
-    uint32 value;
+    uint32_t value;
 
     // Read until we hit limit2.  Except, wait!  limit1 is shorter, so
     // we end up hitting that first, despite having 4 bytes to go on
@@ -1132,7 +1299,7 @@ TEST_F(CodedStreamTest, ExpectAtEnd) {
 
   CodedInputStream::Limit limit = coded_input.PushLimit(4);
 
-  uint32 value;
+  uint32_t value;
   EXPECT_TRUE(coded_input.ReadLittleEndian32(&value));
   EXPECT_TRUE(coded_input.ExpectAtEnd());
 
@@ -1182,27 +1349,25 @@ TEST_F(CodedStreamTest, OverflowLimit) {
 TEST_F(CodedStreamTest, TotalBytesLimit) {
   ArrayInputStream input(buffer_, sizeof(buffer_));
   CodedInputStream coded_input(&input);
-  coded_input.SetTotalBytesLimit(16, -1);
+  coded_input.SetTotalBytesLimit(16);
   EXPECT_EQ(16, coded_input.BytesUntilTotalBytesLimit());
 
   std::string str;
   EXPECT_TRUE(coded_input.ReadString(&str, 16));
   EXPECT_EQ(0, coded_input.BytesUntilTotalBytesLimit());
 
-  std::vector<std::string> errors;
-
   {
-    ScopedMemoryLog error_log;
+    absl::ScopedMockLog log(absl::MockLogDefault::kDisallowUnexpected);
+    EXPECT_CALL(
+        log,
+        Log(absl::LogSeverity::kError, testing::_,
+            testing::HasSubstr(
+                "A protocol message was rejected because it was too big")));
+    log.StartCapturingLogs();
     EXPECT_FALSE(coded_input.ReadString(&str, 1));
-    errors = error_log.GetMessages(ERROR);
   }
 
-  ASSERT_EQ(1, errors.size());
-  EXPECT_PRED_FORMAT2(testing::IsSubstring,
-                      "A protocol message was rejected because it was too big",
-                      errors[0]);
-
-  coded_input.SetTotalBytesLimit(32, -1);
+  coded_input.SetTotalBytesLimit(32);
   EXPECT_EQ(16, coded_input.BytesUntilTotalBytesLimit());
   EXPECT_TRUE(coded_input.ReadString(&str, 16));
   EXPECT_EQ(0, coded_input.BytesUntilTotalBytesLimit());
@@ -1215,7 +1380,7 @@ TEST_F(CodedStreamTest, TotalBytesLimitNotValidMessageEnd) {
   CodedInputStream coded_input(&input);
 
   // Set both total_bytes_limit and a regular limit at 16 bytes.
-  coded_input.SetTotalBytesLimit(16, -1);
+  coded_input.SetTotalBytesLimit(16);
   CodedInputStream::Limit limit = coded_input.PushLimit(16);
 
   // Read 16 bytes.
@@ -1277,7 +1442,6 @@ TEST_F(CodedStreamTest, RecursionLimit) {
 class ReallyBigInputStream : public ZeroCopyInputStream {
  public:
   ReallyBigInputStream() : backup_amount_(0), buffer_count_(0) {}
-  ~ReallyBigInputStream() {}
 
   // implements ZeroCopyInputStream ----------------------------------
   bool Next(const void** data, int* size) override {
@@ -1305,19 +1469,19 @@ class ReallyBigInputStream : public ZeroCopyInputStream {
   void BackUp(int count) override { backup_amount_ = count; }
 
   bool Skip(int count) override {
-    GOOGLE_LOG(FATAL) << "Not implemented.";
+    ABSL_LOG(FATAL) << "Not implemented.";
     return false;
   }
   int64_t ByteCount() const override {
-    GOOGLE_LOG(FATAL) << "Not implemented.";
+    ABSL_LOG(FATAL) << "Not implemented.";
     return 0;
   }
 
   int backup_amount_;
 
  private:
-  char buffer_[1024];
-  int64 buffer_count_;
+  char buffer_[1024] = {};
+  int64_t buffer_count_;
 };
 
 TEST_F(CodedStreamTest, InputOver2G) {
@@ -1325,22 +1489,23 @@ TEST_F(CodedStreamTest, InputOver2G) {
   // input.BackUp() with the correct number of bytes on destruction.
   ReallyBigInputStream input;
 
-  std::vector<std::string> errors;
-
   {
-    ScopedMemoryLog error_log;
+    absl::ScopedMockLog log(absl::MockLogDefault::kDisallowUnexpected);
+    EXPECT_CALL(log, Log(absl::LogSeverity::kError, testing::_, testing::_))
+        .Times(0);
+    log.StartCapturingLogs();
     CodedInputStream coded_input(&input);
     std::string str;
     EXPECT_TRUE(coded_input.ReadString(&str, 512));
     EXPECT_TRUE(coded_input.ReadString(&str, 1024));
-    errors = error_log.GetMessages(ERROR);
   }
 
   EXPECT_EQ(INT_MAX - 512, input.backup_amount_);
-  EXPECT_EQ(0, errors.size());
 }
 
 }  // namespace
 }  // namespace io
 }  // namespace protobuf
 }  // namespace google
+
+#include "google/protobuf/port_undef.inc"
